#ifndef __itkekADMeanSquaresImageToImageMetric_h
#define __itkekADMeanSquaresImageToImageMetric_h

#include "itkImageToImageMetric.h"
#include "itkPoint.h"
#include "itkVectorLinearInterpolateImageFunction.h"
#include "itkVectorNearestNeighborInterpolateImageFunction.h"

#include "itkekADOLCImageValue.h"
#include "adolc.h"


namespace itkek
{

//  using namespace itk;



/** \class ADMeanSquaresImageToImageMetric
 * \brief Computes similarity between two objects to be registered
 *
 * This class is the same as the itkek::MeanSquaresImageToImageMetric
 * except that it computes derivatives using automatic differentiation.
 * Note: It only works if the transform is an itkek::ADRevolveBSplineDeformableTransform.
 *
 * \ingroup RegistrationMetrics
 */
template < class TFixedImage, class TMovingImage > 
class ITK_EXPORT ADMeanSquaresImageToImageMetric : 
	 public itk::ImageToImageMetric< TFixedImage, TMovingImage>
{
public:

  /** Standard class typedefs. */
  typedef ADMeanSquaresImageToImageMetric    Self;
  typedef itk::ImageToImageMetric< TFixedImage, TMovingImage> Superclass;
      
  typedef itk::SmartPointer<Self>         Pointer;
  typedef itk::SmartPointer<const Self>   ConstPointer;

  /** Method for creation through the object factory. */
  itkNewMacro(Self);
 
  /** Run-time type information (and related methods). */
  itkTypeMacro(ADMeanSquaresImageToImageMetric, itk::ImageToImageMetric);


  /** Types transferred from the base class */
  typedef typename Superclass::RealType                 RealType;
  typedef typename Superclass::TransformType            TransformType;
  typedef typename Superclass::TransformPointer         TransformPointer;
  typedef typename Superclass::TransformParametersType  TransformParametersType;
  typedef typename Superclass::TransformJacobianType    TransformJacobianType;
  typedef typename Superclass::GradientPixelType        GradientPixelType;

  typedef typename Superclass::MeasureType              MeasureType;
  typedef typename Superclass::DerivativeType           DerivativeType;
  typedef typename Superclass::FixedImageType           FixedImageType;
  typedef typename Superclass::MovingImageType          MovingImageType;
  typedef typename Superclass::FixedImageConstPointer   FixedImageConstPointer;
  typedef typename Superclass::MovingImageConstPointer  MovingImageConstPointer;



  /**  Get the value for single valued optimizers. */
  MeasureType GetValue( const TransformParametersType & parameters ) const;

  /** Get the derivatives of the match measure. */
  void GetDerivative( const TransformParametersType & parameters,
                            DerivativeType & derivative ) const;

  /**  Get value and derivatives for multiple valued optimizers. */
  void GetValueAndDerivative( const TransformParametersType & parameters,
                              MeasureType& Value, DerivativeType& Derivative ) const;

     

protected:
  ADMeanSquaresImageToImageMetric();
  virtual ~ADMeanSquaresImageToImageMetric() {};

private:
  ADMeanSquaresImageToImageMetric(const Self&); //purposely not implemented
  void operator=(const Self&); //purposely not implemented

//  typedef  itk::VectorLinearInterpolateImageFunction< 
  typedef  itk::VectorNearestNeighborInterpolateImageFunction< 
                                     typename Superclass::GradientImageType,
                                     double >    GradientInterpolatorType;


  mutable ADOLCImageValue<MovingImageType::ImageDimension> m_ADOLCImageValue;


};

} // end namespace itkek

#ifndef ITK_MANUAL_INSTANTIATION
#include "itkekADMeanSquaresImageToImageMetric.txx"
#endif

#endif



