#ifndef __itkekADOLCImageValue_h
#define __itkekADOLCImageValue_h


#include "adolc.h"


namespace itkek
{

// This class is used in automatic differentiation as follows:
// When we've transformed a point in one image to the another
// we need to interpolate it in the other image. Thus the transformed
// is really a function of the entire image. Therefore in theory the
// entire transformed image should be declared as type adouble. To avoid
// we simply approximate the image around the point.


template <unsigned int Dimension>
class ADOLCImageValue
{
};

// Now specialize it for the cases of 2D and 3D

template <>
class ADOLCImageValue<2>
{
public:

  template <class GradientPixelType, class OutputPointType>
  void SetParameters(const double function, 
		     const GradientPixelType& gradient, 
		     const OutputPointType &point)
  {m_a = gradient[0];
  m_b = gradient[1];
  m_d = function - m_a*point[0] - m_b*point[1];}

  adouble &GetImageValue(const adoublev &point)
  {val = m_a*point[0] + m_b*point[1] + m_d;
  return val;}

   ADOLCImageValue():m_a(0.0),m_b(0.0),m_d(0.0),val(0.0){}

private:
  double m_a;
  double m_b;
  double m_d;
  adouble val;

};

template <>
class ADOLCImageValue<3>
{
public:

  template <class GradientPixelType, class OutputPointType>
  void SetParameters(const double function, 
		     const GradientPixelType& gradient, 
		     const OutputPointType &point)
  {m_a = gradient[0];
  m_b = gradient[1];
  m_c = gradient[2];
  m_d = function - m_a*point[0] - m_b*point[1] - m_c*point[2];}
//  {gv = const_cast<double*>(&gradient[0]);
//  m_d = function - gv[0].value()*point[0] - gv[1].value()*point[1] -
//     gv[2].value()*point[2];}

  adouble &GetImageValue(const adoublev &point)
  {val = m_a*point[0] + m_b*point[1] + m_c*point[2] + m_d;
  return val;}

      // The following combines the previous 2 functions into a single
      // function. 
  template <class GradientPixelType>
  adouble &GetImageValue(const adoublev &point, 
			 const GradientPixelType& gradient, 
			 const double func)
  { gv = const_cast<double*>(gradient.GetDataPointer());
    val = point * gv;
    val -= (val.value() - func);
    return val;}

  ADOLCImageValue():m_a(0.0),m_b(0.0),m_c(0.0),m_d(0.0),val(0.0),gv(3){gv=0.0;}

private:
  double m_a;
  double m_b;
  double m_c;
  double m_d;
  adouble val;
  adoublev gv;

};



/*
class ADOLCImageValue
{
public:

  template <class GradientPixelType, class OutputPointType>
  void SetParameters(const double function, 
		     const GradientPixelType& gradient, 
		     const OutputPointType &point)
  {m_a = gradient[0];
  m_b = gradient[1];
  m_c = gradient[2];
  m_d = function - m_a*point[0] - m_b*point[1] - m_c*point[2];}

  adouble &GetImageValue(const adoublev &point)
  {val = m_a*point[0] + m_b*point[1] + m_c*point[2] + m_d;
  return val;}
//      {val = m_d;
//      val += m_a*point[0];
//      val += m_b*point[1];
//      val += m_c*point[2];
//      return val;}

private:
  double m_a;
  double m_b;
  double m_c;
  double m_d;
  adouble val;

};

*/


}


#endif
