#ifndef _itkekADRevolveImageToImageMetricBase_txx
#define _itkekADRevolveImageToImageMetricBase_txx


#include "itkekADRevolveImageToImageMetricBase.h"
#include <time.h>



namespace itkek
{



// define static variables

template <class TFixedImage, class TMovingImage> 
typename ADRevolveImageToImageMetricBase<TFixedImage,TMovingImage>
::FixedImageConstPointer 
ADRevolveImageToImageMetricBase<TFixedImage,TMovingImage>::s_FixedImage;

template <class TFixedImage, class TMovingImage> 
typename ADRevolveImageToImageMetricBase<TFixedImage,TMovingImage>
::MovingImageConstPointer 
ADRevolveImageToImageMetricBase<TFixedImage,TMovingImage>::s_MovingImage;

template <class TFixedImage, class TMovingImage> 
typename ADRevolveImageToImageMetricBase<TFixedImage,TMovingImage>
::InterpolatorPointer 
ADRevolveImageToImageMetricBase<TFixedImage,TMovingImage>::s_Interpolator;

template <class TFixedImage, class TMovingImage> 
typename ADRevolveImageToImageMetricBase<TFixedImage,TMovingImage>
::GradientImagePointer 
ADRevolveImageToImageMetricBase<TFixedImage,TMovingImage>::s_GradientImage;

template <class TFixedImage, class TMovingImage> 
typename ADRevolveImageToImageMetricBase<TFixedImage,TMovingImage>
::GradientInterpolatorPointer 
ADRevolveImageToImageMetricBase<TFixedImage,TMovingImage>
::s_GradientInterpolator;

template <class TFixedImage, class TMovingImage> 
adolc::Revolve 
ADRevolveImageToImageMetricBase<TFixedImage,TMovingImage>::s_Revolve;

template <class TFixedImage, class TMovingImage> 
int
ADRevolveImageToImageMetricBase<TFixedImage,TMovingImage>::s_NumberOfVoxelsInImage;

template <class TFixedImage, class TMovingImage> 
int
ADRevolveImageToImageMetricBase<TFixedImage,TMovingImage>::s_NumberOfVoxelsPerTimeStep;



/*
 * Constructor
 */
template <class TFixedImage, class TMovingImage> 
ADRevolveImageToImageMetricBase<TFixedImage,TMovingImage>
::ADRevolveImageToImageMetricBase()
{
   s_Revolve.SetSnaps(20);
}


/*
 * Destructor
 */
template <class TFixedImage, class TMovingImage> 
ADRevolveImageToImageMetricBase<TFixedImage,TMovingImage>
::~ADRevolveImageToImageMetricBase()
{

}




template <class TFixedImage, class TMovingImage> 
void 
ADRevolveImageToImageMetricBase<TFixedImage,TMovingImage>::
Initialize(void) throw ( itk::ExceptionObject )
{

   this->Superclass::Initialize();



   s_FixedImage = this->m_FixedImage;
   s_MovingImage = this->m_MovingImage;
   s_Interpolator = this->m_Interpolator;
   s_GradientImage = this->m_GradientImage;

   s_GradientInterpolator = GradientInterpolatorType::New();
   s_GradientInterpolator->SetInputImage( s_GradientImage );

   s_NumberOfVoxelsInImage = 
      s_FixedImage->GetLargestPossibleRegion().GetNumberOfPixels();

   s_NumberOfVoxelsPerTimeStep = 500;

   if(m_TransformAD.IsNull())
   {
      itkExceptionMacro(<< "Error: Transform is null ");
   }

   this->AllocateIndependents(this->GetNumberOfParameters());

   

   m_TransformAD->SetParametersAD(m_Independents);
   m_TransformAD->EnableCheckpointing();




   //s_Revolve.SetSnaps(64);
   s_Revolve.SetFunction(this);

//   PRINT(s_Revolve.GetSnaps());

#ifdef IRAD_VERBOSE_OUTPUT
   s_Revolve.SetInfo(1);
//   s_Revolve.SetInfo(2);
#else
   s_Revolve.SetInfo(0);
#endif


   //PRINT(s_NumberOfVoxelsInImage);
   //PRINT(s_NumberOfVoxelsPerTimeStep);
   //PRINT(this->GetNumberOfParameters());
}


template <class TFixedImage, class TMovingImage> 
void
ADRevolveImageToImageMetricBase<TFixedImage,TMovingImage>
::SetTransform (TransformType* _arg) 
{
   if (this->m_TransformAD != _arg) 
   { 

      this->m_Transform = _arg; 
      this->Modified();

      // dynamic cast to adTransform
      typename ADRevolveTransformType::Pointer adTransform =
		  dynamic_cast< ADRevolveTransformType* >(_arg);
	    
      if(adTransform)
      {
		  m_TransformAD = adTransform; 
      }
      else
      {
		  itkExceptionMacro(<< "Error: Could not dynamic cast to "
							"ADRevolveTransformBase.");
      }

   } 
}


template <class TFixedImage, class TMovingImage> 
void
ADRevolveImageToImageMetricBase<TFixedImage,TMovingImage>
::CopyData(const ObjectiveFunctionBase* rhs)
{
   if(this == rhs)
      return;

   this->ObjectiveFunctionBase::CopyData(rhs);


   // First dynamic cast rhs to Self
   const Self* rhs_cast =
      dynamic_cast< const Self* >(rhs);


   if(!rhs_cast)
   {
      itkExceptionMacro(<< "Error: Could not cast to "
	 "ADRevolveImageToImageMetric.");
   }


   m_TransformAD->CopyData(rhs_cast->m_TransformAD);


}


/*
 * Get the match Measure
 */
template <class TFixedImage, class TMovingImage> 
typename ADRevolveImageToImageMetricBase<
   TFixedImage,TMovingImage>::MeasureType
ADRevolveImageToImageMetricBase<TFixedImage,TMovingImage>
::GetValue( const TransformParametersType & parameters ) const
{

   PRINT(" Starting function evaluation");
   //std::cout << " Starting function evaluation\n";


   MeasureType measure = 0.0;


   //EVAL( unsigned int start_time2 = time(0) );
   EVAL( unsigned int start_time = clock() );



   s_Revolve.Evaluate(parameters.data_block(),&measure,0);



   EVAL( unsigned int end_time = clock() );
   //EVAL( unsigned int end_time2 = time(0) );


//   PRINT(parameters);
//   PRINT(measure);

   EVAL( double elapsed_time = (end_time - start_time )/(double)CLOCKS_PER_SEC );
   PRINT(" Run time for one function evaluation (using clock())    =\n" 
	 << elapsed_time       << "  sec  =  "
	 << (elapsed_time/60.0) << "  min  =  "
	 << (elapsed_time/3600.0) << "  hr");
   PRINT("IJ Run_time_for_function "  << elapsed_time);
//   PRINT("IJ Run_time_for_one_function_evaluation_in_seconds "  << elapsed_time);
//    std::cout << " Run time for one function evaluation (using clock())    =\n" 
// 	     << elapsed_time       << "  sec  =  ";
//    std::cout << (elapsed_time/60.0) << "  min  =  ";
//    std::cout << (elapsed_time/3600.0) << "  hr" << std::endl;

//   double elapsed_time2 = (double)(end_time2 - start_time2 );
//    std::cout << " Run time for one function evaluation (using time(0))    =\n" 
// 	     << elapsed_time2       << "  sec  =  ";
//    std::cout << (elapsed_time2/60.0) << "  min  =  ";
//    std::cout << (elapsed_time2/3600.0) << "  hr\n\n" << std::endl;

//    PRINTNL("IJ parameters_for_function ");
//    EVAL(for(int i=0;i<this->GetNumberOfParameters();i++))
//       PRINTNL(parameters[i] << " ");
//    PRINT("");



   return measure;
}


/*
 * Get the Derivative Measure
 */
template < class TFixedImage, class TMovingImage>
void
ADRevolveImageToImageMetricBase<TFixedImage,TMovingImage>
::GetDerivative( const TransformParametersType & parameters,
		 DerivativeType & derivative  ) const
{
   MeasureType measure = 0.0;
   this->GetValueAndDerivative(parameters,measure,derivative);
}



/*
 * Get both the match Measure and the Derivative Measure 
 */
template <class TFixedImage, class TMovingImage> 
void
ADRevolveImageToImageMetricBase<TFixedImage,TMovingImage>
::GetValueAndDerivative(const TransformParametersType& parameters, 
			MeasureType& measure, 
			DerivativeType& derivative) const
{

   PRINT(" Starting function+gradient evaluation");
   //std::cout << " Starting function+gradient evaluation\n";


   derivative = DerivativeType( parameters.size() );
   derivative.Fill(itk::NumericTraits<typename DerivativeType::ValueType>::Zero);



   //EVAL( unsigned int start_time2 = time(0) );
   EVAL( unsigned int start_time = clock() );



   s_Revolve.Evaluate(parameters.data_block(),&measure,derivative.data_block());



   EVAL( unsigned int end_time = clock() );
   //EVAL( unsigned int end_time2 = time(0) );


//   PRINT(parameters);
//   PRINT(derivative);
//   PRINT(measure);

   EVAL( double elapsed_time = (end_time - start_time )/(double)CLOCKS_PER_SEC );
   PRINT(" Run time for one function evaluation and its "
	 << "gradient (using clock())    =\n" 
	 << elapsed_time       << "  sec  =  "
	 << (elapsed_time/60.0) << "  min  =  "
	 << (elapsed_time/3600.0) << "  hr");
   PRINT("IJ Run_time_for_function_and_gradient " << elapsed_time);
//    std::cout << " Run time for one function evaluation and its "
// 	     << "gradient (using clock())    =\n" 
// 	     << elapsed_time       << "  sec  =  ";
//    std::cout << (elapsed_time/60.0) << "  min  =  ";
//    std::cout << (elapsed_time/3600.0) << "  hr" << std::endl;

//    double elapsed_time2 = (double)(end_time2 - start_time2 );
//    std::cout << " Run time for one function evaluation and its " 
// 	     << "gradient (using time(0))    =\n" 
// 	     << elapsed_time2       << "  sec  =  ";
//    std::cout << (elapsed_time2/60.0) << "  min  =  ";
//    std::cout << (elapsed_time2/3600.0) << "  hr\n\n" << std::endl;

//    PRINTNL("IJ parameters_for_function_and_gradient ");
//    EVAL(for(int i=0;i<this->GetNumberOfParameters();i++))
//       PRINTNL(parameters[i] << " ");
//    PRINT("");
//    PRINTNL("IJ derivative_for_function_and_gradient ");
//    EVAL(for(int i=0;i<this->GetNumberOfParameters();i++))
//       PRINTNL(derivative[i] << " ");
//    PRINT("");



}


}



#endif
