#ifndef __itkekADRevolveMeanSquaresImageToImageMetric_h
#define __itkekADRevolveMeanSquaresImageToImageMetric_h

#include "itkekADRevolveImageToImageMetricBase.h"

#include "itkPoint.h"
#include "itkVector.h"
#include "itkPointSet.h"
#include "itkLinearInterpolateImageFunction.h"
#include "itkCovariantVector.h"
#include "itkImage.h"
#include "itkVectorLinearInterpolateImageFunction.h"
#include "itkCentralDifferenceImageFunction.h"
#include "vnl/vnl_matrix.h"
#include "vnl/vnl_matrix_fixed.h"

#include <vector>
#include <map>


namespace itkek
{



/**
   This class represents the objective function used for non-rigid
   registration. 
 */
class ITK_EXPORT ADRevolveMeanSquaresImageToImageMetric : 
	 public ADRevolveImageToImageMetricBase<itk::Image<float, 3u>,
						itk::Image<float, 3u> >
{

   public:

      typedef ADRevolveMeanSquaresImageToImageMetric    Self;
      typedef ADRevolveImageToImageMetricBase<itk::Image<float, 3u>,
					      itk::Image<float, 3u> > Superclass;
      
      typedef itk::SmartPointer<Self>         Pointer;
      typedef itk::SmartPointer<const Self>   ConstPointer;

      /** Method for creation through the object factory. */
      itkNewMacro(Self);

      /** Run-time type information (and related methods). */
      itkTypeMacro(ADRevolveMeanSquaresImageToImageMetric,
		   ADRevolveImageToImageMetricBase);


      typedef itk::Point<double,3> PointType;
      typedef std::vector<PointType> PointsContainer;
      typedef itk::Vector<double,3> VectorType;

      typedef std::vector<std::pair<int,
				    VectorType> > FixedDisplacementsContainer;

      typedef itk::Image< float,3 > MovingImageType ;
      typedef MovingImageType::ConstPointer MovingImagePointer;
      typedef itk::Image< float,3 > FixedImageType ;
      typedef FixedImageType::ConstPointer FixedImagePointer;

      typedef itk::LinearInterpolateImageFunction<
	 MovingImageType, double > 	InterpolatorType;
      typedef InterpolatorType::Pointer 	MovingImageInterpolatorPointer;



      void Initialize(void) throw ( itk::ExceptionObject );

   protected:

      ADRevolveMeanSquaresImageToImageMetric() {}
      virtual ~ADRevolveMeanSquaresImageToImageMetric() {}


   private:


      typedef PointType InputPointType;
      typedef PointType OutputPointType;
      typedef itk::Vector<double,3> InputVectorType;
      typedef itk::Vector<double,3> OutputVectorType;



      typedef adoublev ADVectorType;


      typedef itk::CovariantVector< double,3 >  GradientPixelType;
      typedef itk::Image< GradientPixelType,3 > GradientImageType ;
      typedef itk::SmartPointer< GradientImageType > GradientImagePointer;


      typedef  itk::VectorLinearInterpolateImageFunction<
         GradientImageType, double >    GradientInterpolatorType;
      typedef GradientInterpolatorType::Pointer GradientInterpolatorPointer;




      // The main function.
      virtual void Function(Label& label);

      virtual adolc::ObjectiveFunctionBase* Create() const;

      virtual LabelsContainerType GetLabels() const;

      virtual void CopyData(const ObjectiveFunctionBase* rhs);




      void SSD(Label& label);




      ADRevolveMeanSquaresImageToImageMetric(
	 const ADRevolveMeanSquaresImageToImageMetric& rhs); // not implemented
      ADRevolveMeanSquaresImageToImageMetric& operator=(
	 const ADRevolveMeanSquaresImageToImageMetric& rhs); //not implemented



      //Local data
      int m_Ints[10];

};

}


#endif
