#ifndef _itkekLoadSaveAffineTransform_txx
#define _itkekLoadSaveAffineTransform_txx


#include "itkekLoadSaveAffineTransform.h"
#include "itkekMacro.h"

#include "itkImageRegionConstIterator.h"
#include "itkNumericTraits.h"
#include "itkAffineTransform.h"
#include "itkCastImageFilter.h"
#include <fstream>



namespace itkek
{




// This function loads an affine transform and stores it in a
// a AffineTransform pointer provided by the user

   template <class AffineTransformSmartPointer>
   int LoadSaveAffineTransform::LoadAffineTransform(const char* const filename, 
				  const AffineTransformSmartPointer transform, 
				  const bool invert)
   {

     std::ifstream transformfile(filename);
     if(!transformfile) 
     {
	itkGenericExceptionMacro(<< "Error: cannot open file");
     }


     if(transform.IsNull())
     {
	itkGenericExceptionMacro(<< "Error: pointer has not been initialized");
     }

     
     typedef typename AffineTransformSmartPointer::ObjectType AffineTransformType;



     typename AffineTransformType::ParametersType parameters(16);
     parameters.Fill(-1.0);


     for(int i=0;i<16;i++)
     {
	transformfile >> parameters[i];
     }

     // First Check to make sure the last line is 0 0 0 1
     if(parameters[12] != 0.0 || parameters[13] != 0.0 ||
	parameters[14] != 0.0 || parameters[15] != 1.0)
     {
	itkGenericExceptionMacro(<< "Error reading file: The transform does not appear to "
	   "be an affine transform.");
     }

     transformfile.close();

     typename AffineTransformType::MatrixType matrix;
     typename AffineTransformType::OffsetType offset;

     int c=0;
     for(int i=0; i < 3; i++)
     {
	for(int j=0; j < 3; j++)
	{
	   matrix[i][j] = parameters[c++];
	}
	offset[i] = parameters[c++];
     }



     transform->SetMatrix(matrix);
     transform->SetOffset(offset);

     if(invert == true )
     {
	
	typename AffineTransformType::Pointer inverse = AffineTransformType::New();
	transform->GetInverse(inverse);
	
	transform->SetMatrix(inverse->GetMatrix());
	transform->SetOffset(inverse->GetOffset());

	
     }


     return 0;

   }


// This function saves an affine transform
   template <class AffineTransformSmartPointer>
   int LoadSaveAffineTransform::SaveAffineTransform(const char* const  filename, 
				  const AffineTransformSmartPointer transform)
   {

     std::ofstream transformfile(filename);
     if(!transformfile)
     {
	itkGenericExceptionMacro(<< "Error: cannot open file");
     }


     if(transform.IsNull())
     {
	itkGenericExceptionMacro(<< "Error: pointer has not been initialized");
     }


     transformfile.setf(std::ios_base::scientific,std::ios_base::floatfield);
     transformfile.precision(16);


     typedef typename AffineTransformSmartPointer::ObjectType AffineTransformType;


     typename AffineTransformType::MatrixType matrix = transform->GetMatrix();
     typename AffineTransformType::OffsetType offset = transform->GetOffset();


     for(int i=0; i < 3; i++)
     {
	for(int j=0; j < 3; j++)
	{
	   transformfile << matrix[i][j] << "  ";
	}
	transformfile << offset[i] << "\n";
     }


     transformfile << "0  0  0  1\n";

     transformfile.close();

     return 0;

   }



} // end namespace itkek

#endif
