#ifndef __rbfmatrix_h
#define __rbfmatrix_h

#include <vector>
#include <cmath>
#include <cassert>
#include <iostream>


namespace itkek
{


// This matrix represents the square rbf operator matrix
// To use this class one setup function must be called:
// SetupMatrix(...)

// This matrix is templated over the point type used
// for setting up the matrix

// This matrix doesn't actually store its components;
// instead when a method is called requesting data, 
// the data is generated on the fly
template <class TPoint>
class rbfmatrix
{
   public:
   
      typedef double ScalarType;
      typedef TPoint PointType;
      typedef std::vector<PointType> PointSetType;
   

      // Set the matrix: 
      void SetupMatrix(const PointSetType& landmarks,
		       int k, ScalarType t)
      {
	 m_Tau2 = t*t; 
	 m_K = k; 

	 if(m_K <= 0)
	 {
	    itkGenericExceptionMacro(<< "m_K must be positive");
	 }

	 m_SourceLandmarks = landmarks; 
	 m_NumberOfPoints = m_SourceLandmarks.size();
	 assert(m_NumberOfPoints > 0);

	 num_rows = m_NumberOfPoints;

      }

	  
      ScalarType operator()(int i,int j) const
      {

	 assert(i >= 0);
	 assert(i < num_rows);
	 assert(j >= 0);
	 assert(j < num_rows);
      

	 return this->BasicFunction(m_SourceLandmarks[i], 
				    m_SourceLandmarks[j]);
	 
      }
   
    
      int size1() const { return num_rows; }
      int size2() const { return num_rows; }
      int rows() const { return num_rows; }
      int cols() const { return num_rows; }

      
   
      ScalarType BasicFunction(const PointType & point, 
			       const PointType & center) const
      {
	 const ScalarType x = point[0] - center[0];
	 const ScalarType y = point[1] - center[1];
	 const ScalarType z = point[2] - center[2];
	 return sqrt( std::pow(x*x + y*y + z*z + m_Tau2, m_K ) );
      }

   private:

      // The list of source landmarks
      PointSetType m_SourceLandmarks;

      int num_rows; // since matrix is square there is no num_cols variable
      int m_NumberOfPoints;
      ScalarType m_Tau2;
      int m_K;
   
};
   
}

#endif
